unit Appl;

interface

uses
  Windows, Forms, SysUtils, Classes, Controls, IniFiles;

function FindPrevInstance(Name: string): boolean;
procedure StopPrevInstance;
function ShowPrevInstance(Name: string): boolean;
function FindAndShowPrevInstance(Name: string): boolean;
function ReadFormRes(ResName: string; List: TStringList): boolean;
function FindForm(FormName: string): TCustomForm;

function RtmPPI(FormName: string): integer;
function DsgnPPI(FormName: string): integer;
function ScaleValue(Value, NewPPI, OldPPI: Integer): integer;
procedure ChangeConstraints(Control: TControl; NewPPI, OldPPI: integer);

function GetIniFileName: string;
function GetSectionName(Component: TComponent): string;
procedure WriteIniShowCmd(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure ReadIniShowCmd(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure WriteIniFlags(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure ReadIniFlags(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure WriteIniWidth(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; NewPPI, OldPPI: integer);
procedure ReadIniWidth(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; List: TStringList; NewPPI, OldPPI: integer);
procedure WriteIniHeight(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; NewPPI, OldPPI: integer);
procedure ReadIniHeight(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; List: TStringList; NewPPI, OldPPI: integer);
procedure WriteIniLeft(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure ReadIniLeft(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure WriteIniTop(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure ReadIniTop(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
procedure SaveFormState(Form: TForm);
procedure LoadFormState(Form: TForm);

const
  strNil = '';
  strDelphiMagicText = '0';
  strResTextHeight = 'TextHeight';
  strResClientWidth = 'ClientWidth';
  strResClientHeight = 'ClientHeight';
  cfgFileIniExt = '.ini';
  cfgFormState = 'State';
  cfgFormFlags = 'Flags';
  cfgFormLeft = 'Left';
  cfgFormTop = 'Top';
  cfgFormWidth = 'Width';
  cfgFormHeight = 'Height';

var
  Mutex: THandle = 0;

implementation

function FindPrevInstance(Name: string): boolean;
var
  Temp: THandle;
begin
  Temp := CreateMutex(nil, False, PChar(Name));
  Result := (GetLastError = ERROR_ALREADY_EXISTS);
  if Result then
    CloseHandle(Temp)
  else
    Mutex := Temp;
end;

procedure StopPrevInstance;
begin
  if Mutex > 0 then
    CloseHandle(Mutex);
end;

function ShowPrevInstance(Name: string): boolean;
var
  PrevInstance: HWND;
begin
  Result := False;
  PrevInstance := FindWindow('TApplication', PChar(Name));
  if PrevInstance <> 0 then
  begin
    if IsIconic(PrevInstance) then
      ShowWindow(PrevInstance, SW_RESTORE);
    SetForegroundWindow(PrevInstance);
    Result := True;
  end;
end;

function FindAndShowPrevInstance(Name: string): boolean;
begin
  Result := FindPrevInstance(Name);
  if Result then
    ShowPrevInstance(Name);
end;

function ReadFormRes(ResName: string; List: TStringList): boolean;
var
  Prop, Value: string;
  Stream: TResourceStream;
  Reader: TReader;
  HRsrc: THandle;
begin
  List.Clear;
  HRsrc := FindResource(HInstance, PChar(ResName), RT_RCDATA);
  Result := HRsrc <> 0;
  if not Result then Exit;
  Stream := TResourceStream.Create(HInstance, ResName, RT_RCDATA);
  Reader := TReader.Create(Stream, 4096);
  try
    Reader.ReadSignature;
    Reader.ReadStr;
    Reader.ReadStr;
    while not Reader.EndOfList do
    begin
      Prop := Reader.ReadStr;
      Value := strNil;
      case Reader.NextValue of
        vaInt8, vaInt16, vaInt32:
          Value := IntToStr(Reader.ReadInteger);
        vaString:
          Value := Reader.ReadString;
        else
          Reader.SkipValue;
      end;
      if Value <> strNil then
        List.Add(Format('%s=%s',[Prop,Value]));
    end;
    Reader.CheckValue(vaNull);
  finally
    Reader.Free;
    Stream.Free;
  end;
end;

function FindForm(FormName: string): TCustomForm;
var
  I: integer;
begin
  Result := nil;
  for I := 0 to Screen.FormCount - 1 do
    if Screen.Forms[I].Name = FormName then
    begin
      Result := Screen.Forms[I];
      Break;
    end;
end;

function RtmPPI(FormName: string): integer;
var
  Form: TCustomForm;
begin
  Result := 0;
  Form := FindForm(FormName);
  if Form <> nil then
    Result := Form.Canvas.TextHeight(strDelphiMagicText);
end;

function DsgnPPI(FormName: string): integer;
var
  List: TStringList;
  Form: TCustomForm;
begin
  List := TStringList.Create;
  try
    Form := FindForm(FormName);
    if Form <> nil then
    begin
      ReadFormRes(Form.ClassName, List);
      Result := StrToInt(List.Values[strResTextHeight]);
    end;
  finally
    List.Free;
  end;
end;

function ScaleValue(Value, NewPPI, OldPPI: integer): integer;
begin
  Result := MulDiv(Value, NewPPI, OldPPI);
end;

procedure ChangeConstraints(Control: TControl; NewPPI, OldPPI: integer);
begin
  with Control.Constraints do
  begin
    if NewPPI > OldPPI then
    begin
      if MaxHeight > 0 then MaxHeight := ScaleValue(MaxHeight, NewPPI, OldPPI);
      if MinHeight > 0 then MinHeight := ScaleValue(MinHeight, NewPPI, OldPPI);
      if MaxWidth > 0 then MaxWidth := ScaleValue(MaxWidth, NewPPI, OldPPI);
      if MinWidth > 0 then MinWidth := ScaleValue(MinWidth, NewPPI, OldPPI);
    end
    else
    begin
      if MinHeight > 0 then MinHeight := ScaleValue(MinHeight, NewPPI, OldPPI);
      if MaxHeight > 0 then MaxHeight := ScaleValue(MaxHeight, NewPPI, OldPPI);
      if MinWidth > 0 then MinWidth := ScaleValue(MinWidth, NewPPI, OldPPI);
      if MaxWidth > 0 then MaxWidth := ScaleValue(MaxWidth, NewPPI, OldPPI);
    end;
  end;
end;

function GetIniFileName: string;
begin
  Result := ChangeFileExt(Application.ExeName, cfgFileIniExt);
end;

function GetSectionName(Component: TComponent): string;
begin
  Result := Component.Name;
end;

procedure WriteIniShowCmd(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
begin
  Ini.WriteInteger(Section, cfgFormState, Place.ShowCmd);
end;

procedure ReadIniShowCmd(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
begin
  Place.ShowCmd := Ini.ReadInteger(Section, cfgFormState, 1);
end;

procedure WriteIniFlags(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
begin
  Ini.WriteInteger(Section, cfgFormFlags, Place.Flags);
end;

procedure ReadIniFlags(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
begin
  Place.Flags := Ini.ReadInteger(Section, cfgFormFlags, 0);
end;

procedure WriteIniWidth(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; NewPPI, OldPPI: integer);
var
  FormW, BorderW, ClientW: integer;
begin
  FormW := Place.rcNormalPosition.Right - Place.rcNormalPosition.Left;
  BorderW := Form.Width - Form.ClientWidth;
  ClientW := MulDiv(FormW - BorderW, OldPPI, NewPPI);
  Ini.WriteInteger(Section, cfgFormWidth, ClientW);
end;

procedure ReadIniWidth(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; List: TStringList; NewPPI, OldPPI: integer);
var
  DefaultW, BorderW, ClientW: integer;
begin
  DefaultW := StrToInt(List.Values[strResClientWidth]);
  if DefaultW <> 0 then
  begin
    BorderW := Form.Width - Form.ClientWidth;
    ClientW := MulDiv(Ini.ReadInteger(Section, cfgFormWidth, DefaultW), NewPPI, OldPPI);
    Place.rcNormalPosition.Right := Place.rcNormalPosition.Left + BorderW + ClientW;
  end;
end;

procedure WriteIniHeight(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; NewPPI, OldPPI: integer);
var
  FormH, BorderH, ClientH: integer;
begin
  FormH := Place.rcNormalPosition.Bottom - Place.rcNormalPosition.Top;
  BorderH := Form.Height - Form.ClientHeight;
  ClientH := MulDiv(FormH - BorderH, OldPPI, NewPPI);
  Ini.WriteInteger(Section, cfgFormHeight, ClientH);
end;

procedure ReadIniHeight(Form: TForm; Ini: TIniFile; var Section: string;
  var Place: TWindowPlacement; List: TStringList; NewPPI, OldPPI: integer);
var
  DefaultH, BorderH, ClientH: integer;
begin
  DefaultH := StrToInt(List.Values[strResClientHeight]);
  if DefaultH <> 0 then
  begin
    BorderH := Form.Height - Form.ClientHeight;
    ClientH := MulDiv(Ini.ReadInteger(Section, cfgFormHeight, DefaultH), NewPPI, OldPPI);
    Place.rcNormalPosition.Bottom := Place.rcNormalPosition.Top + BorderH + ClientH;
  end;
end;

procedure WriteIniLeft(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
begin
  Ini.WriteInteger(Section, cfgFormLeft, Place.rcNormalPosition.Left);
end;

procedure ReadIniLeft(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
var
  FormW, DefaultL: integer;
begin
  FormW := Place.rcNormalPosition.Right - Place.rcNormalPosition.Left;
  DefaultL := (Screen.Width - FormW) div 2;
  Place.rcNormalPosition.Left := Ini.ReadInteger(Section,
    cfgFormLeft, DefaultL);
  Place.rcNormalPosition.Right := Place.rcNormalPosition.Left + FormW;
end;

procedure WriteIniTop(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
begin
  Ini.WriteInteger(Section, cfgFormTop, Place.rcNormalPosition.Top);
end;

procedure ReadIniTop(Ini: TIniFile; var Section: string; var Place: TWindowPlacement);
var
  FormH, DefaultT: integer;
begin
  FormH := Place.rcNormalPosition.Bottom - Place.rcNormalPosition.Top;
  DefaultT := (Screen.Height - FormH) div 2;
  Place.rcNormalPosition.Top := Ini.ReadInteger(Section,
    cfgFormTop, DefaultT);
  Place.rcNormalPosition.Bottom := Place.rcNormalPosition.Top + FormH;
end;

procedure ScaleFormConstraints(Form: TForm; NewPPI, OldPPI: integer);
var
  MaxH, MinH, MaxW, MinW, BorderW, BorderH: integer;
begin
  BorderW := Form.Width - Form.ClientWidth;
  BorderH := Form.Height - Form.ClientHeight;
  with Form.Constraints do
  begin
    if MaxHeight > 0 then
      MaxH := ScaleValue(MaxHeight - BorderH, NewPPI, OldPPI) + BorderH;
    if MinHeight > 0 then
      MinH := ScaleValue(MinHeight - BorderH, NewPPI, OldPPI) + BorderH;
    if MaxWidth > 0 then
      MaxW := ScaleValue(MaxWidth - BorderW, NewPPI, OldPPI) + BorderW;
    if MinWidth > 0 then
      MinW := ScaleValue(MinWidth - BorderW, NewPPI, OldPPI) + BorderW;
    if NewPPI > OldPPI then
    begin
      if MaxHeight > 0 then MaxHeight := MaxH;
      if MinHeight > 0 then MinHeight := MinH;
      if MaxWidth > 0 then MaxWidth := MaxW;
      if MinWidth > 0 then MinWidth := MinW;
    end
    else
    begin
      if MinHeight > 0 then MinHeight := MinH;
      if MaxHeight > 0 then MaxHeight := MaxH;
      if MinWidth > 0 then MinWidth := MinW;
      if MaxWidth > 0 then MaxWidth := MaxW;
    end;
  end;
end;

procedure SaveFormState(Form: TForm);
var
  NewPPI, OldPPI: integer;
  Place: TWindowPlacement;
  Section: string;
  Ini: TIniFile;
begin
  OldPPI := RtmPPI(Form.Name);
  NewPPI := DsgnPPI(Form.Name);
  if (OldPPI > 0) and (NewPPI > 0) then
  begin
    Place.Length := SizeOf(Place);
    if GetWindowPlacement(Form.Handle, @Place) then
    begin
      Section := GetSectionName(Form);
      Ini := TIniFile.Create(GetIniFileName);
      try
        WriteIniShowCmd(Ini, Section, Place);
        WriteIniFlags(Ini, Section, Place);
        WriteIniLeft(Ini, Section, Place);
        WriteIniTop(Ini, Section, Place);
        WriteIniWidth(Form, Ini, Section, Place, NewPPI, OldPPI);
        WriteIniHeight(Form, Ini, Section, Place, NewPPI, OldPPI);
      finally
        Ini.Free;
      end;
    end;
  end;
end;

procedure LoadFormState(Form: TForm);
var
  NewPPI, OldPPI: integer;
  Place: TWindowPlacement;
  Section: string;
  Ini: TIniFile;
  List: TStringList;
begin
  OldPPI := RtmPPI(Form.Name);
  NewPPI := DsgnPPI(Form.Name);
  if (OldPPI > 0) and (NewPPI > 0) then
  begin
    Place.Length := SizeOf(Place);
    if GetWindowPlacement(Form.Handle, @Place) then
    begin
      Section := GetSectionName(Form);
      Ini := TIniFile.Create(GetIniFileName);
      List := TStringList.Create;
      try
        ReadFormRes(Form.ClassName, List);
        ScaleFormConstraints(Form, NewPPI, OldPPI);
        ReadIniWidth(Form, Ini, Section, Place, List, NewPPI, OldPPI);
        ReadIniLeft(Ini, Section, Place);
        ReadIniHeight(Form, Ini, Section, Place, List, NewPPI, OldPPI);
        ReadIniTop(Ini, Section, Place);
        ReadIniShowCmd(Ini, Section, Place);
        ReadIniFlags(Ini, Section, Place);
        SetWindowPlacement(Form.Handle, @Place);
      finally
        List.Free;
        Ini.Free;
      end;
    end;
  end;
end;

initialization
//  if FindAndShowPrevInstance('DUnit') then
//    Halt;
finalization
//  StopPrevInstance;
end.

